#include "cora_fit.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <gsl_routines.h>

static double hlik;
static int index;

/*
 *  determine error in amplitude 'eramp'
 *  for which second derivative 'like2' is used
 *  dellik = 1 for accuracy of error within 68.3%
 *  (actually dellik=1/2, but this turns to 1 because
 *   -2log(lik) is used.)     
 */
void Amperr(Vector *amp)
{
  int i, j, k, nrot;
  double dellik = 1.0;
  Vector source, dummy;
  
  gsl_vector *d,*r;
  gsl_matrix *aerr, *v;

  vec_init (&source);
  vec_init (&dummy);
  vec_alloc(&source,count);
  vec_alloc(&dummy, (nlines < count) ? count : nlines);

  dummy.len = nlines;
  for (i = 0; i < count; ++i)
    {
      for (j=0; j<nlines; ++j) 
	dummy.vec[j] = amp->vec[j]*yc[j][i];
      source.vec[i] = bbb.vec[i]+total(&dummy);
    }

#if defined(OBSOLETE_CODE)
  dummy.len = count;
  
  for(j=0;j<nlines;++j)
    {
      for(i=0;i<count;++i) 
	dummy.vec[i] = ncount.vec[i]*yc[j][i]*yc[j][i]/
	  (source.vec[i]*source.vec[i]);
      like2          = 2.*total(&dummy);   
      /* comment: 2nd derivative of likelihood */ 
      eramp.vec[j]   = sqrt(2.*dellik/like2);
      if(finite(eramp.vec[i]) == 0)
	{
	  eramp.vec[i] = -1.;
	  printf("*  Warning  *: calculated NaN for error in amplitude(%d)\n",
		 i);
	}
    }
#endif

  /*   calculate all second derivatives of likelihood  
   */
  d    = gsl_vector_alloc(nlines);
  r    = gsl_vector_alloc(nlines);
  v    = gsl_matrix_alloc(nlines,nlines);
  aerr = gsl_matrix_alloc(nlines,nlines);
 
  for(j=0; j<nlines; j++)
    {
      for(k=0; k<nlines; k++) 
	aerr->data[j * aerr->tda + k] = 0.0;
    }

  for(j=0; j<nlines; ++j)
    {
      for(i=0; i<count; ++i)
	{
	  aerr->data[j * aerr->tda + j] += ncount.vec[i]*yc[j][i]*yc[j][i]/
	    (source.vec[i]*source.vec[i]);
	  for(k = j+1; k < nlines; ++k) 
	    {
	      aerr->data[j * aerr->tda + k] += ncount.vec[i]*yc[j][i]*yc[k][i]/
		(source.vec[i]*source.vec[i]);
	      aerr->data[k * aerr->tda + j] =aerr->data[j * aerr->tda + k];
	    }
	}
    }
 
  gsl_eigen_jacobi_impl(aerr, d, v, 800, &nrot);
  gsl_eigen_sort_impl(d, v, GSL_EIGEN_SORT_ABSVALUE);
 
  for(i=0; i<nlines; ++i)
    {
      eramp.vec[i]=0.;
      for(j=0; j<nlines; ++j) 
	eramp.vec[i] += dellik*v->data[i*v->tda + j]*v->data[i*v->tda + j]/d->data[j * d->stride];
      eramp.vec[i] = sqrt(eramp.vec[i]);
    }

  if(verbose)
    printf("Message      : Finished error analysis for Amp\n");

  return;
}

/*  ------------------------ */

double errlik(double x)
     /*      called from 'para_err':
	     determine difference of likelihood compared to hlik
	     The root finder 'zeroin' finds the value x
	     where likelihood=hlik=min(likelihood)+dL               */
     
     /* Perr  =[lam(0)..lam(nlines-1),sig(0)..sig(nlines-1)]
	vector with fit results, i.e. d(Lik)/d(Perr)=0
	index indicates for which variable and for which line
	the error is to be estimated
	x     is the iterated variable provided from 'zeroin'
	hlik  = min(likelihood)+dL, dL depending on the
	desired accuray for the error (c.f. para_err).   */
     
     /*  since only one argument can be provided from
	 functions called from internal routines (zeroin)
	 index and hlik are provided by global variables */
     
{
  double lik0 = 0., test;
  
  test          = mP.vec[index];
  mP.vec[index] = x;
  lik0          = funcname(&(mP.vec[0]),&mP.len) - hlik;
  mP.vec[index] = test;
  return(lik0);
}

/*  ------------------------ */

void para_err()
     /*   determine error flag_err in vec with numerical method
	  flag     : =0 error for lambda, =1 for sigma
	  delta=1. : returns error with 68.3% confidence
		     delta=0.2514 returns error with 10% confidence
	  conftransf: transforms errors to higher confidence level
	  dvec     : shift in vec to get two other points on parabola
	  find point where likelihood value reaches minimum value + delta:
	  lik1=llik-delta			lik0=llik+delta
	  \			  /
	  \			/
	  --   llik  ---
	  
	  testm  	        vec(lline)         testp
	  this is done numerically with zeroin procedure
	  
	  flag indicates whether errors for sigma or lambda are to be found
	  I must admit this is a failed attempt to keep this routine flexible 
     */
{

  int i, j, k, fitnum, ibord=0;
  double llik, delta=.0158, dL=0.3, dS=0.1, tol=1e-4;

  double conftransf = 1./0.1257;
  Vector test1, test2, P, vec_min, vec_plu;
  double testm, testp;
  
  /*  store variables being changed in error analysis 
   */
  double *ycback[MAXLINES],lik0,lik1;
  Vector amp,lam_back,sig_back;
  
  vec_init(&amp);
  vec_init(&lam_back);
  vec_init(&sig_back);

  vec_alloc(&amp,      nlines);
  vec_alloc(&lam_back, nlines);
  vec_alloc(&sig_back, nlines);

  for(i=0; i<nlines; ++i)
    {
      ycback[i]=malloc(count*sizeof(double));
      if (ycback[i] == NULL) 
	{
	  fprintf (stderr, "*  Warning  *: Out of memory\n"); 
	  exit (8);
	}
    }

  for(i=0; i<nlines; ++i)
    {
      if(finite(llambda.vec[i])==0 || finite(lsigma.vec[i])==0)
       {
	 printf (" ** ERROR ** No error analysis possible"\
		 "for NaN fit results\n");
	 return;
       }
      amp.vec[i]      = ampout.vec[i];
      lam_back.vec[i] = llambda.vec[i];
      sig_back.vec[i] = lsigma.vec[i];
      for (j=0; j<count; ++j) 
	ycback[i][j]=yc[i][j];
    }
  /*              end store  	                   */

  
  vec_init(&P);
  vec_init(&test1);
  vec_init(&test2);
  vec_init(&vec_min);
  vec_init(&vec_plu);
  vec_alloc(&P,2*nlines);
  vec_alloc(&test1,nlines);
  vec_alloc(&test2,nlines);
  vec_alloc(&vec_min,nlines);
  vec_alloc(&vec_plu,nlines);

  /*  determine hlik, the by delta higher likelihood value  
   */
  llik = funcname(&(mP.vec[0]),&mP.len);

  hlik = llik + delta;

  /*    <<<<<  Lambda  >>>>>   */

  if(ilamit>0)
    {
      if (verbose)
         fprintf(stdout, "Message      : calculating errors for lambda\n");

      if(idlam == 0)
	ibord=nlines;
      else
	ibord=1;  

     /* find point where root is included: 
        lik0 > hlik > like < hlik < lik1  
     */

      /* make backup of lambda
       */
      k = 0;
      for(i=0; i<nlines; ++i)
	{
	  if (fit_wave[i] == TRUE && fit_lock[i] == FALSE)
	    {
	      test1.vec[k] = lam_back.vec[i];
	      test2.vec[k] = lam_back.vec[i];
	      ++k;
	    }
	}
      fitnum = k;


      /* determine points left/right such that like[left]/like[right]
       * is higher by One+ than like[bestfit]
       */
      for(i=0; i < ibord; ++i) 
	{
	  k = 0;
	  if (fit_wave[i] == TRUE && fit_lock[i] == FALSE)
	    {
	      testm = dL;
	      lik0  = hlik-delta;    /* make while condition TRUE  */
	      lik1  = hlik-delta;
	      
	      while(lik0<hlik || lik1<hlik)
		{
		  test1.vec[k] = lam_back.vec[i]-testm*lsigma.vec[i];
		  lik0         = func_lamlik(&(test1.vec[0]), &fitnum);
		  
		  test2.vec[k] = lam_back.vec[i]+testm*lsigma.vec[i];
		  lik1         = func_lamlik(&(test2.vec[0]), &fitnum);
		  
		  if (lik0<hlik || lik1<hlik) 
		    testm = 2.0 * testm;
		}
	      vec_min.vec[i] = test1.vec[k];
	      vec_plu.vec[i] = test2.vec[k];
	      test1.vec[k]   = lam_back.vec[i];
	      test2.vec[k]   = lam_back.vec[i];
	      llambda.vec[i] = lam_back.vec[i];
	      ++k;
	    }
	}

      /*   find points where likelihood value reaches minimum value+delta:  
       */
      for(i=0; i<ibord; ++i) 
       {
	  if (fit_wave[i] == TRUE && fit_lock[i] == FALSE)
	    {
	      index = i;    /*  index of changed variable */
	      testm = zeroin(vec_min.vec[i], llambda.vec[i], errlik, tol);
	      llambda.vec[i] = lam_back.vec[i];
	      testp = zeroin(llambda.vec[i], vec_plu.vec[i], errlik, tol);
	      llambda.vec[i] = lam_back.vec[i];
	      
	      if(llambda.vec[i]-testm>testp-llambda.vec[i])
		erlam.vec[i] = conftransf * (llambda.vec[i]-testm);
	      else
		erlam.vec[i] = conftransf * (testp-llambda.vec[i]);

	      /* FIXME why use startP here ???
	       */
	      if (startP.vec[i]-dlam >= llambda.vec[i]-erlam.vec[i] ||
		  startP.vec[i]+dlam <= llambda.vec[i]+erlam.vec[i])
		{
		  printf("*  Warning  *: Cannot calculate error for "\
			 "lambda(%d): error (%f) out of range (%f)\n",
			 i, erlam.vec[i], dlam);
		  erlam.vec[i] = -1.;
		}
	    }
       }

      /* if the multiplet is locked, there is only one free wavelength,
       * hence only one error.
       */
      if(idlam > 0) 
	for(i=1; i<nlines; ++i) 
	  erlam.vec[i] = erlam.vec[0];

      for(i=0; i < nlines; ++i)
	if (fit_wave[i] == FALSE)
	  erlam.vec[i] = (-1.0);

      for(i=1; i < nlines; ++i)
	if (fit_lock[i] == TRUE)
	  erlam.vec[i] = erlam.vec[i-1];
      
    }

  /*    <<<<<  Sigma  >>>>>   */

  if (isigit > 0)
    {
      /* find point where root is included
	 lik0 > hlik > like < hlik < lik1  */
      
      if (verbose)
	fprintf(stdout, "Message      : calculating errors for sigma\n");

      lik0 = hlik - delta;
      lik1 = hlik - delta;

      k = 0;
      for(i=0; i < nlines; ++i)
	{
	  if (fit_sigm[i] == TRUE)
	    {
	      test1.vec[k] = sig_back.vec[i];
	      test2.vec[k] = sig_back.vec[i];
	      ++k;
	    }
	}
      fitnum = k;

      for(i = 0; i < nlines; ++i)
	{
	  k = 0;

	  if (fit_sigm[i] == TRUE)
	    {

	      testm = dS;
	      lik0  = hlik-delta;
	      lik1  = hlik-delta;
  

	      while (lik0<hlik || lik1<hlik)
		{
		  test1.vec[k] = sig_back.vec[i] - testm*sig_back.vec[i];
		  lik0         = func_siglik(&(test1.vec[0]), &fitnum);
		  
		  test2.vec[k] = sig_back.vec[i] + testm*sig_back.vec[i];
		  lik1         = func_siglik(&(test2.vec[0]), &fitnum);
		  
		  if(lik0<hlik || lik1<hlik) 
		    testm = 2.0 * testm;

		}

	      vec_min.vec[i] = test1.vec[k];
	      vec_plu.vec[i] = test2.vec[k];
	      test1.vec[k]   = sig_back.vec[i];
	      test2.vec[k]   = sig_back.vec[i];
	      lsigma.vec[i]  = sig_back.vec[i];
	      ++k;
	    }     
	}
      
      /*   find points where likelihood value reaches minimum value+delta:  
       */
      for(i=0; i<nlines; ++i)
	{
	  if (fit_sigm[i] == TRUE)
	    {
	      index = i+ibord;    /*  index of changed variable */
	      testm = zeroin(vec_min.vec[i], lsigma.vec[i], errlik, tol);
	      lsigma.vec[i] = sig_back.vec[i];
	      testp = zeroin(lsigma.vec[i], vec_plu.vec[i], errlik, tol);
	      lsigma.vec[i] = sig_back.vec[i];
	  
	      if( lsigma.vec[i]-testm  >  testp-lsigma.vec[i] )
		ersig.vec[i] = conftransf * (lsigma.vec[i]-testm);
	      else
		ersig.vec[i] = conftransf * (testp-lsigma.vec[i]);
	  
	      /* FIXME why use startP here ???
	       */
	      if (startP.vec[nlines+i]-dsig >= lsigma.vec[i]-ersig.vec[i] ||
		  startP.vec[nlines+i]+dsig <= lsigma.vec[i]+ersig.vec[i])
		{
		  printf("*  Warning  *: Cannot calculate error for "\
			 "sigma(%d): error (%f) out of range (%f)\n",
			 i, ersig.vec[i], dsig);
		  ersig.vec[i] = -1.;
		}
	    }
	}
    }

  for(i=0; i < nlines; ++i)
    if (fit_sigm[i] == FALSE)
      ersig.vec[i] = (-1.0);


  /*   restore changed fit results  
   */
  for(i=0; i<nlines; ++i)
    {
      ampout.vec[i] = amp.vec[i];
      for (j=0; j<count; ++j) 
	yc[i][j]=ycback[i][j];
    }
  /*         end restore           */
    
  return;
}


void checkerr()
{

  Vector Ptest;
  double like,ltest,back;
  int i,ibord=0;
  
  vec_init(&Ptest);
  vec_alloc(&Ptest,mP.len);
  
  for(i=0; i<mP.len; ++i)
    {
      Ptest.vec[i] = mP.vec[i];
    }
  
  like = funcname(&(Ptest.vec[0]), &Ptest.len);
  
  if(ilamit>0)
    {
      if (idlam == 0)
	ibord = nlines;
      else
	ibord = 1;

      if (verbose)
	fprintf(stdout, "Message      : <<< check errors for lambda >>>\n");

      for(i=0; i < ibord; ++i)
	{
	  back = llambda.vec[i];
	  Ptest.vec[i] = back-erlam.vec[i];
	  ltest = funcname(&(Ptest.vec[0]), &Ptest.len);
	  if (verbose)
	    fprintf(stdout, "Message      : %d: ---> %f\n",
		    i, ltest-like);
	  Ptest.vec[i] = back+erlam.vec[i];
	  ltest = funcname(&(Ptest.vec[0]), &Ptest.len);
	  if (verbose)
	    fprintf(stdout, "Message      : %d: +++> %f\n",
		    i,ltest-like);
	  Ptest.vec[i] = back;
	  llambda.vec[i] = back;
	}
       if(verbose) 
	 printf("Message      : Finished error analysis for Lambda\n");
    }

  if (isigit > 0)
    {
      if (verbose)
	fprintf(stdout, "Message      : <<< check errors for sigma >>>\n\n");

      if(ilamit > 0)
        {
          if(idlam == 0)
	    ibord=nlines;
          else
	    ibord=1;
	}

      for(i=ibord; i < ibord+nlines; ++i)
	{
	  back = lsigma.vec[i-ibord];
	  Ptest.vec[i] = back-ersig.vec[i-ibord];
	  ltest = funcname(&(Ptest.vec[0]),&Ptest.len);
	  if (verbose)
	    fprintf(stdout, "Message      : %d: ---> %f\n",i-ibord,ltest-like);
	  Ptest.vec[i] = back+ersig.vec[i-ibord];
	  ltest = funcname((&Ptest.vec[0]),&Ptest.len);
	  if (verbose)
	    fprintf(stdout, "Message      : %d: +++> %f\n",i-ibord,ltest-like);
	  Ptest.vec[i] = back;
	}
      if(verbose) printf("Message      : Finished error analysis for Sigma\n");
    } 
  return;
}
